import logging

from rich.logging import RichHandler

from clarifai.utils.logging import _get_library_name, get_logger


def test_get_logger():
  logger = get_logger("DEBUG", "test_logger")
  assert logger.level == logging.DEBUG
  assert logger.name == "test_logger"
  assert isinstance(logger.handlers[0], RichHandler)


def test_get_logger_defaults():
  logger = get_logger()
  assert logger.level == logging.NOTSET
  assert logger.name == _get_library_name()
  assert isinstance(logger.handlers[0], RichHandler)



# Copyright 2023 Clarifai, Inc.
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""
Triton python backend inference model controller.
"""

import inspect
import os
import shutil
import zipfile
from dataclasses import asdict
from pathlib import Path
from typing import Iterable, Literal, Union

import yaml
from tqdm import tqdm

from ..constants import BUILT_MODEL_EXT
from ..model_config import MODEL_TYPES, ModelConfigClass, base, get_model_config, load_user_config
from ..model_config.base import *  # noqa
from ..model_config.config import parse_config
from ..model_config.triton.serializer import Serializer


def __parse_type_to_class():
  _t = {}
  _classes = inspect.getmembers(base, inspect.isclass)
  for cls_name, cls_obj in _classes:
    if cls_obj.__base__ is base._BaseClarifaiModel:
      _t.update({cls_obj._config.clarifai_model.type: cls_name})
  return _t


_TYPE_TO_CLASS = __parse_type_to_class()


def _get_static_file_path(relative_path: str):
  curr_dir = os.path.dirname(__file__)
  return os.path.join(curr_dir, "static_files", relative_path)


def _read_static_file(relative_path: str):
  path = _get_static_file_path(relative_path)
  with open(path, "r") as f:
    return f.read()


def copy_folder(src_folder, dest_folder, exclude_items=None):

  if exclude_items is None:
    exclude_items = set()

  # Ensure the destination folder exists
  if not os.path.exists(dest_folder):
    os.makedirs(dest_folder)

  loader = tqdm(os.listdir(src_folder))
  if exclude_items:
    print(f"NOTE: skipping {exclude_items}")

  for item in loader:
    loader.set_description(f"copying {item}...")
    src_item = os.path.join(src_folder, item)
    dest_item = os.path.join(dest_folder, item)

    # Skip items in the exclude list
    if item in exclude_items or item.endswith(BUILT_MODEL_EXT):
      continue

    # Copy files directly
    if os.path.isfile(src_item):
      shutil.copy2(src_item, dest_item)

    # Copy directories using copytree
    elif os.path.isdir(src_item):
      shutil.copytree(src_item, dest_item, symlinks=False, ignore=None, dirs_exist_ok=True)


def zip_dir(input: Union[Path, str], zip_filename: Union[Path, str]):
  """
  Zip folder without compressing
  """
  # Convert to Path object
  dir = Path(input)

  with zipfile.ZipFile(zip_filename, "w", zipfile.ZIP_STORED) as zip_file:
    for entry in dir.rglob("*"):
      zip_file.write(entry, entry.relative_to(dir))


class RepositoryBuilder:

  @staticmethod
  def init_repository(model_type: str, working_dir: str, backend=Literal['triton'], **kwargs):
    assert model_type in MODEL_TYPES
    model_type = model_type
    default_model_type_config: ModelConfigClass = get_model_config(model_type)

    os.makedirs(working_dir, exist_ok=True)

    def __write_to(filename, data):
      with open(os.path.join(working_dir, filename), "w") as f:
        f.write(data)

    # create inference.py
    _filename = "inference.py"
    inference_py = _read_static_file(_filename)
    inference_py = inference_py.replace("InferenceModel()",
                                        f"InferenceModel({_TYPE_TO_CLASS[model_type]})")
    inference_py = inference_py.replace("predict_docstring",
                                        eval(_TYPE_TO_CLASS[model_type]).predict.__doc__)
    # create config
    config = asdict(default_model_type_config)
    if backend == "triton":
      max_batch_size = kwargs.get("max_batch_size", None)
      image_shape = kwargs.get("image_shape", None)
      if max_batch_size:
        config['serving_backend']['triton']['max_batch_size'] = max_batch_size
      if image_shape:
        config['serving_backend']['triton']['image_shape'] = image_shape
      config = parse_config(config).dump_to_user_config()
      config_data = yaml.dump(config)
      sample_yaml = _read_static_file("sample_clarifai_config.yaml")
      config_data = sample_yaml + "\n\n" + config_data
      __write_to("clarifai_config.yaml", config_data)
      #
    # create inference.py after checking all configs
    __write_to(_filename, inference_py)
    # create test.py
    __write_to("test.py", _read_static_file("test.py"))
    # create requirements.txt
    __write_to("requirements.txt", _read_static_file("_requirements.txt"))

  @staticmethod
  def build(working_dir: str, output_dir: str = None, name: str = None, backend=Literal['triton']):
    if not output_dir:
      output_dir = working_dir
    else:
      os.makedirs(output_dir, exist_ok=True)

    temp_folder = os.path.join(working_dir, ".cache")
    os.makedirs(temp_folder, exist_ok=True)

    user_config_file = os.path.join(working_dir, "clarifai_config.yaml")
    assert os.path.exists(
        user_config_file
    ), f"FileNotFound: please make sure `clarifai_config.yaml` exists in {working_dir}"
    user_config = load_user_config(user_config_file)

    if backend == "triton":
      triton_1_ver = os.path.join(temp_folder, "1")
      os.makedirs(triton_1_ver, exist_ok=True)
      # check if labels exists
      for output_config in user_config.serving_backend.triton.output:
        if output_config.label_filename:
          user_labels = user_config.clarifai_model.labels
          assert user_labels, f"Model type `{user_config.clarifai_model.type}` requires labels, "\
          f"but can not found value of `clarifai_model.labels` in {user_config_file}. Please update this attribute to build the model"
          with open(os.path.join(temp_folder, "labels.txt"), "w") as f:
            if not isinstance(user_labels, Iterable):
              user_labels = [user_labels]
            f.write("\n".join([str(lb) for lb in user_labels]) + "\n")

      # copy model.py
      shutil.copy(_get_static_file_path("triton/model.py"), triton_1_ver)
      # copy requirements.txt
      shutil.copy(os.path.join(working_dir, "requirements.txt"), temp_folder)
      # copy all other files
      copy_folder(
          working_dir, triton_1_ver, exclude_items=["requirements.txt", ".cache", "__pycache__"])
      # generate config.pbtxt
      _config_pbtxt_serializer = Serializer(user_config.serving_backend.triton)
      _config_pbtxt_serializer.to_file(temp_folder)

    else:
      raise ValueError(f"backend must be ['triton'], got {backend}")

    clarifai_model_name = name or user_config.clarifai_model.clarifai_model_id or "model"
    clarifai_model_name += BUILT_MODEL_EXT
    clarifai_model_name = os.path.join(output_dir, clarifai_model_name)

    print(
        "Model building in progress; the duration may vary depending on the size of checkpoints/assets..."
    )
    zip_dir(temp_folder, clarifai_model_name)
    print(f"Finished. Your model is located at {clarifai_model_name}")

    return clarifai_model_name



import unittest

from clarifai.models.model_serving.repo_build import BaseTest


class CustomTest(unittest.TestCase):
  """
  BaseTest loads the InferenceModel from the inference.py file in the current working directory.
  To execute the predict method of the InferenceModel, use the predict method in BaseTest.
  It takes the exact same inputs and inference parameters, returning the same outputs as InferenceModel.predict.
  The difference is that BaseTest.predict verifies your_infer_parameters against config.clarifai_models.inference_parameters and checks the output values.

  For example, test input value of visual-classifier

  def test_input(self):
    import cv2
    path = "path/to/image"
    img = cv2.imread(path)
    outputs = self.model.predict([img], infer_param1=..., infer_param2=...)
    print(outputs)
    assert outputs

  """

  def setUp(self) -> None:
    your_infer_parameter = dict(
    )  # for example dict(float_var=0.12, string_var="test", _secret_string_var="secret")
    self.model = BaseTest(your_infer_parameter)

  def test_default_cases(self):
    """Test your model with dummy inputs.
    In general, you only need to run this test to check your InferneceModel implementation.
    In case the default inputs makes your model failed for some reason (not because of assert in `test_with_default_inputs`),
    you can comment out this test.
    """
    self.model.test_with_default_inputs()

  def test_specific_case1(self):
    """ Implement your test case"""
    pass



import logging
import os
import uuid

import pytest

from clarifai.client.app import App
from clarifai.client.user import User
from clarifai.constants.search import DEFAULT_TOP_K

NOW = uuid.uuid4().hex[:10]
MAIN_APP_ID = "main"
MAIN_APP_USER_ID = "clarifai"
GENERAL_MODEL_ID = "general-image-recognition"
General_Workflow_ID = "General"

CREATE_APP_USER_ID = os.environ["CLARIFAI_USER_ID"]
CREATE_APP_ID = f"ci_test_app_{NOW}"
CREATE_MODEL_ID = f"ci_test_model_{NOW}"
CREATE_DATASET_ID = f"ci_test_dataset_{NOW}"
CREATE_MODULE_ID = f"ci_test_module_{NOW}"
CREATE_RUNNER_ID = f"ci_test_runner_{NOW}"

CLARIFAI_PAT = os.environ["CLARIFAI_PAT"]


@pytest.fixture
def create_app():
  return App(user_id=CREATE_APP_USER_ID, app_id=CREATE_APP_ID, pat=CLARIFAI_PAT)


@pytest.fixture
def app():
  return App(user_id=MAIN_APP_USER_ID, app_id=MAIN_APP_ID, pat=CLARIFAI_PAT)


@pytest.fixture
def client():
  return User(user_id=MAIN_APP_USER_ID, pat=CLARIFAI_PAT)


@pytest.mark.requires_secrets
class TestApp:
  """Tests for the App class and its methods.

    CRUD operations are tested for each of the following resources:
    - app
    - dataset
    - model
    - workflow

    Note: Update to be added later.
    """

  def test_list_models(self, app):
    all_models = list(app.list_models(page_no=1))
    assert len(all_models) == 16  #default per_page is 16

  def test_list_workflows(self, app):
    all_workflows = list(app.list_workflows(page_no=1, per_page=10))
    assert len(all_workflows) == 10

  def test_list_apps(self, client):
    all_apps = list(client.list_apps())
    assert len(all_apps) > 0

  def test_get_model(self, client):
    model = client.app(app_id=MAIN_APP_ID).model(model_id=GENERAL_MODEL_ID)
    versions = list(model.list_versions())
    assert len(versions) == 2  #test for list_versions
    assert model.id == GENERAL_MODEL_ID and model.app_id == MAIN_APP_ID and model.user_id == MAIN_APP_USER_ID

  def test_get_workflow(self, client):
    workflow = client.app(app_id=MAIN_APP_ID).workflow(workflow_id=General_Workflow_ID)
    versions = list(workflow.list_versions())
    assert len(versions) == 1  #test for list_versions
    assert workflow.id == General_Workflow_ID and workflow.app_id == MAIN_APP_ID and workflow.user_id == MAIN_APP_USER_ID

  def test_create_app(self):
    app = User(user_id=CREATE_APP_USER_ID, pat=CLARIFAI_PAT).create_app(app_id=CREATE_APP_ID)
    assert app.id == CREATE_APP_ID and app.user_id == CREATE_APP_USER_ID

  def test_create_search(self, create_app):
    search = create_app.search()
    assert search.top_k == DEFAULT_TOP_K and search.metric_distance == "EUCLIDEAN_DISTANCE"

  def test_create_dataset(self, create_app):
    dataset = create_app.create_dataset(CREATE_DATASET_ID)
    assert dataset.id == CREATE_DATASET_ID and dataset.app_id == CREATE_APP_ID and dataset.user_id == CREATE_APP_USER_ID

  def test_create_model(self, create_app):
    model = create_app.create_model(CREATE_MODEL_ID)
    assert model.id == CREATE_MODEL_ID and model.app_id == CREATE_APP_ID and model.user_id == CREATE_APP_USER_ID

  def test_create_module(self, create_app):
    module = create_app.create_module(CREATE_MODULE_ID, description="CI test module")
    assert module.id == CREATE_MODULE_ID and module.app_id == CREATE_APP_ID and module.user_id == CREATE_APP_USER_ID

  def test_create_runner(self, client):
    client = User(user_id=CREATE_APP_USER_ID, pat=CLARIFAI_PAT)
    runner_info = client.create_runner(
        CREATE_RUNNER_ID, labels=["ci runner"], description="CI test runner")
    assert runner_info.get("runner_id") == CREATE_RUNNER_ID and runner_info.get(
        "user_id") == CREATE_APP_USER_ID

  def test_get_dataset(self, create_app):
    dataset = create_app.dataset(dataset_id=CREATE_DATASET_ID)
    versions = list(dataset.list_versions())
    assert len(versions) == 0  #test for list_versions
    assert dataset.id == CREATE_DATASET_ID and dataset.app_id == CREATE_APP_ID and dataset.user_id == CREATE_APP_USER_ID

  def test_list_datasets(self, create_app):
    all_datasets = list(create_app.list_datasets())
    assert len(all_datasets) == 1

  def test_delete_dataset(self, create_app, caplog):
    with caplog.at_level(logging.INFO):
      create_app.delete_dataset(CREATE_DATASET_ID)
      assert "SUCCESS" in caplog.text

  def test_delete_model(self, create_app, caplog):
    with caplog.at_level(logging.INFO):
      create_app.delete_model(CREATE_MODEL_ID)
      assert "SUCCESS" in caplog.text

  def test_delete_module(self, create_app, caplog):
    with caplog.at_level(logging.INFO):
      create_app.delete_module(CREATE_MODULE_ID)
      assert "SUCCESS" in caplog.text

  def test_delete_runner(self, caplog):
    client = User(user_id=CREATE_APP_USER_ID)
    with caplog.at_level(logging.INFO):
      client.delete_runner(CREATE_RUNNER_ID)
      assert "SUCCESS" in caplog.text

  def test_delete_app(self, caplog):
    with caplog.at_level(logging.INFO):
      User(user_id=CREATE_APP_USER_ID).delete_app(CREATE_APP_ID)
      assert "SUCCESS" in caplog.text
